#ifndef _AVOID_OFFSIDE_CONSTRAINT_H_
#define _AVOID_OFFSIDE_CONSTRAINT_H_

#include "Common.h"
#include "Constraint.h"

const int AVOID_OFFSIDE_CONSTRAINT_ID = 602;

/**
  Nebenbedingung, die Abseitsentscheidungen verhindern soll.
  
  Abseits wird vom Agenten an zwei Stellen ber\"ucksichtigt. Der Spieler
  versucht generell, nicht im passiven Abseits zu stehen, und er wird keinen
  eigenen Spieler, der im Abseits steht, anspielen. Die dritte, bisher nicht
  ber\"ucksichtigte Situation ist, da\3 ein im Abseits stehender Spieler, dem
  Ball nicht n\"aher als 9.15m kommen darf. Dazu mu\3 aber im Weltmodell bekannt
  sein, wer den Ball zuletzt ber\"uhrt hat. Das ist schwer festzustellen.

  {\small {\bf Quelltext}

  [\URL[Header]{src/AvoidOffsideConstraint.h} | [\URL[Implementierung]{src/AvoidOffsideConstraint.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: AvoidOffsideConstraint.h,v $
  *   $Revision: 1.2 $
  *   $Date: 1998/08/14 13:19:58 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class AvoidOffsideConstraint : public Constraint
{
public:
  /** Konstruktor, der die Id setzt. */
  AvoidOffsideConstraint();

  /** Beschreibung als String fuer die Ausgabe auf einem Terminal. */
  virtual String toString() const;

  /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. */
  virtual StorableObject* copy() const;

  /** Virtueller Destruktor */
  virtual ~AvoidOffsideConstraint();

protected:
  /** Wenn sich der Agent im Abseits befindet, wir maxAvoidOffsideUtility
  geliefert, sonst 0. */
  virtual int computeExpectedUtility() const;
};

  
#endif
