
/*
 * Aktueller Stand
 *
 *   $RCSfile: CollectDataOption.cc,v $
 *   $Revision: 1.18 $
 *   $Date: 1998/08/14 13:19:59 $
 *   $Author: myritz $ 
 *
 */
 
#include "CollectDataOption.h"
#include "Constraints.h"
#include "WorldModel.h"
#include "Skills.h"

CollectDataOption::CollectDataOption() : PassiveOption()
{
  theId = COLLECT_DATA_OPTION_ID;
}  


int CollectDataOption::baseUtility() const 
{
  INFO( Log::PASSIVE_OPTIONS, 2, "ENTER CollectDataOption::baseUtility\n" );
  ModelledBall ball = globalWorldModel->ball();
  const ModelledPlayer& player = globalWorldModel->player();
  int utility = PassiveOption::baseUtility();

  if( utility == NEUTRAL_UTILITY ) { // keine Daten "uber mich
    if( globalWorldModel->gameState() != WorldModel::BEFORE_KICK_OFF &&
        globalWorldModel->gameState() != WorldModel::GOAL ) {
      // ich wei"s nicht wo ich bin und kann nicht beamen
      utility = MAX_UTILITY;
      setTurnToAngle( player.faceDir() - PI/2 );
    }
  }
  else if( utility == MAX_UTILITY ) { // Daten "uber mich, und weg vom Ball
    utility = NEUTRAL_UTILITY;
    if( ball.posReliability().get() < Reliability::RANDOM ) {
      ball.simulateSteps();
      if( player.isInFieldOfView( ball ) &&
          globalWorldModel->decisionReason() == 
          WorldModel::DECISION_REASON_SEE_INFO ) {
        // genaue Daten "uber Ball -> andere Daten sammeln
        utility = globalConf->minCollectDataUtility;
        theTurnReason = TURN_REASON_COLLECT_OTHER_DATA;
        setTurnToAngle( player.faceDir() + PI );
      }
      else {
        // keine genauen Daten "uber Ball -> zum Ball drehen
        utility = globalConf->normalCollectDataUtility;
        theTurnReason = TURN_REASON_GET_NEW_BALL_INFO;
        setTurnToAngle( ( ball.pos() - player.pos() ).angle() );
      }
    }
    else { // Ballinformationen sind inkorekt
      utility = globalConf->maxCollectDataUtility;
      theTurnReason = TURN_REASON_SEEK_BALL;
      setTurnToAngle( player.faceDir() + player.visibleAngle() );
    }
  }
  else { // Daten "uber mich und am Ball
    // doNothing
  }

  INFO( Log::PASSIVE_OPTIONS, 6, "utility: " + String( utility ) + "\n" );
  INFO( Log::PASSIVE_OPTIONS, 2, "LEAVE CollectDataOption::baseUtility\n" );

  return utility;
}

int CollectDataOption::constraintCost( const Constraint *constraint ) const 
{
  int cost;
  switch( constraint->id() ) {
  case CONSERVE_STAMINA_CONSTRAINT_ID:
    cost = NEUTRAL_COST;
    break;
  case AVOID_OFFSIDE_CONSTRAINT_ID:
    cost = MAX_COST;
    break;
  default:
    cost = PassiveOption::constraintCost( constraint );
  }
  return cost;
}

void CollectDataOption::adaptToPreviousIntention( 
  const Option *previousIntention ) 
{
}

Plan CollectDataOption::generatePlan() const 
{
  Plan plan;
  if( !isTargetComputed() ) {
    throwException( Exception( "Target should had been already computed" ) );
  }
  globalSkills->resetSituation();
  switch( theTurnReason ) {
    case TURN_REASON_SEEK_BALL:
    case TURN_REASON_COLLECT_OTHER_DATA:
      plan.addPlan( globalSkills->setViewWidth( VIEW_WIDTH_WIDE ) );
      break;
    case TURN_REASON_GET_NEW_BALL_INFO:
      plan.addPlan( globalSkills->setViewWidth( VIEW_WIDTH_NORMAL ) );
    default:
      break;
  }
  // Kommunikation !!
  plan.addPlan( isSomethingToSay() );
  plan.addPlan( globalSkills->turnTo( turnToAngle() ) );
  plan.addPlan( globalSkills->doNothing() );
  setIsPlanCompleted( YES );
  return plan;
}


StorableObject* CollectDataOption::copy() const 
{
  return new CollectDataOption( *this );
}

String CollectDataOption::toString() const
{
  String output;
  output = String( "( CollectDataOption" ) +
    String( " utility: " ) + String( expectedUtility() ) + 
    String( " turnToAngle: " );
  if( isTargetComputed() ) {
    output = output + String( degFromRadiant( turnToAngle() ) );
  }
  else {
    output = output + "(not fixed)";
  }
  output = output + 
    String( " contTime: " ) + String( forceContinuationTime() );
  output = output + String( " ) " );

  return output;
}




 
