#ifndef _CONTROL_BALL_OPTION_H_
#define _CONTROL_BALL_OPTION_H_

#include "KickOption.h"
#include "Field.h"

const int CONTROL_BALL_OPTION_ID = 708;


/**
  Diese Klasse steht fuer den Wunsch, die alleinige Ballkontrolle zu erlangen.
  
  Dabei wird versucht, den eigenen Koerper zwischen Ball und Gegner zu bewegen.

  {\small {\bf Quelltext}

  [\URL[Header]{src/ControlBallOption.h} | [\URL[Implementierung]{src/ControlBallOption.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: ControlBallOption.h,v $
  *   $Revision: 1.3 $
  *   $Date: 1998/08/14 13:20:12 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class ControlBallOption : public KickOption
{
public:
  /** Erzeugt eine ControlBallOption. */
  ControlBallOption();

  /** Beschreibung als String fuer die Ausgabe auf einem Terminal. */
  virtual String toString() const;

  /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. Diese Methode muss fuer JEDE Unterklasse neu definiert 
      werden. */
  virtual StorableObject* copy() const;
  
  /** Prueft, ob die Absicht tatsaechlich umsetzbar ist.
    @return YES oder NO. */
  virtual bool isPossible() const;

  /** Erzeugt einen Plan zur Umsetzung der Absicht. 
    Zur Erzeugung wird der Skill \Ref{gainBallControl}() aufgerufen.
    @return Plan, der die ersten Schritte zur Umsetzung der Absicht enthaelt.
    Die Zahl der Schritte laesst sich konfigurieren (Variable maxLengthOfPlan
    in agent.conf). */
  virtual Plan generatePlan() const;
  
protected:  
  /** Der Grundnutzen des Wunsches. Dieser Teil der Nutzenberechnung ist nur 
      vom Weltmodell abhaengig.
      @return Im Allgemeinen ein Wert zwischen MAX_UTILITY und MIN_UTILITY. */
  int baseUtility() const;

  /** Gibt an, ob ein anderer Spieler auch den Ball im Kontrollradius hat. */
  mutable bool theIsBallInRangeOfOtherPlayer;
};

#endif



