#ifndef _COLLECT_DATA_OPTION_H_
#define _COLLECT_DATA_OPTION_H_

#include "PassiveOption.h"

const int COLLECT_DATA_OPTION_ID = 702;


/**
  Diese Klasse steht fuer den Wunsch, Daten \"uber die Umwelt zu sammeln.
  Diese Option hat immer einen Nutzen, auch wenn er nicht so hoch ist. Wenn
  die Verl\"a\3lichkeit des Balles sinkt erh\"oht sich der Nutzen dieser 
  Option.
  Bei Auswahl des Wunsches dreht sich der Spieler um Daten \"uber die Umwelt
  zu sammeln.

  {\small {\bf Quelltext}

  [\URL[Header]{src/CollectDataOption.h} | [\URL[Implementierung]{src/CollectDataOption.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: CollectDataOption.h,v $
  *   $Revision: 1.8 $
  *   $Date: 1998/08/14 13:20:00 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class CollectDataOption : public PassiveOption
{
public:

  /** @name M\"ogliche Gr\"unde f\"ur diese Option. */
  //@{
    /// 
    const int TURN_REASON_MIN = 0;
    ///
    const int TURN_REASON_COLLECT_OTHER_DATA = 0;
    ///
    const int TURN_REASON_GET_NEW_BALL_INFO = 1;
    ///
    const int TURN_REASON_SEEK_BALL = 2;
    ///
    const int TURN_REASON_MAX = 2;
  //@}

  /** Erzeugt eine CollectDataOption. */
  CollectDataOption();
  
  /** Anpassung an die bisherige Absicht. */
  void adaptToPreviousIntention( const Option *previousIntention );

  /** Erzeugt einen Plan zur Umsetzung der Absicht.  */
  Plan generatePlan() const;

  /** Beschreibung als String fuer die Ausgabe auf einem Terminal. */
  virtual String toString() const;

  /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. Diese Methode muss fuer JEDE Unterklasse neu definiert 
      werden. */
  virtual StorableObject* copy() const;
    
private:
  
  /** Der Grundnutzen des Wunsches. Der Nutzen des Wunsches steigt stark,
   wenn die Informationen \"uber den Ball veraltert sind. */
  int baseUtility() const;

  /** Liefert die Kosten, die durch die Einhaltung einer Nebenbedingung   
      entstehen.
      @param constraint Zeiger auf die einzuhaltende Nebenbedingung
      @return Ein Wert zwischen NEUTRAL_COST und MAX_COST. */
  int constraintCost( const Constraint *constraint ) const;

  /** Setzt das Drehziel. (Die Richtung in die sich der Spieler drehen
    soll). */
  void setTurnToAngle( double aTurnToAngle ) const;
  
  /** Erfragt das Drehziel. (Die Richtung in die sich der Spieler drehen
    soll). */
  double turnToAngle() const;

  /** Das Drehziel. (Die Richtung in die sich der Spieler drehen soll). */
  mutable double theTurnToAngle;

  /** Der Drehgrund. Warum der Agent sich drehen soll. */
  mutable int theTurnReason;
};

inline void CollectDataOption::setTurnToAngle( double aTurnToAngle ) const
{
  theTurnToAngle = aTurnToAngle;
  setIsTargetComputed( YES );
}

inline double CollectDataOption::turnToAngle() const
{
  return theTurnToAngle;
}

#endif

