
/*
 * Aktueller Stand
 *
 *   $RCSfile: ForwardPassOption.cc,v $
 *
 *   $Revision: 1.11 $
 *   $Date: 1998/08/14 13:20:33 $
 *   $Author: myritz $ 
 *
 */
 
#include "ForwardPassOption.h"
#include "Constraints.h"
#include "WorldModel.h"
#include "Skills.h"
#include "math.h"
#include "Reliability.h"
#include "InvalidIdException.h"
#include "StraightLine.h"

ForwardPassOption::ForwardPassOption() : KickOption() 
{
  theId = FORWARD_PASS_OPTION_ID;
}  

int ForwardPassOption::baseUtility() const 
{

  INFO( Log::ACTIVE_OPTIONS, 2, 
      "ENTER ForwardPassOption::baseUtility\n" );

  int bestUtility = KickOption::baseUtility();
  if ( bestUtility == MAX_UTILITY ) {
    // Es liegt eine Kick-Situation vor. Jetzt genaue Werte ausrechnen.
    
    // Szenario vorbereiten
    Situation scenario( globalWorldModel->situation() );
    ModelledBall ball = scenario.ball();
    ModelledPlayer player = scenario.playerWithId( globalWorldModel->id() );
    ModelledPlayer tempPlayer = 
        scenario.playerWithId( globalWorldModel->id() );
    Vector speed;
    bool wasSimulationSuccessful;

    // Standardabweichung bei jedem Simulationsschritt in Rad ( 10 Grad )
    double standardDifferenceAngle = radFromDegree( 10 );
    // temporaere Abweichung beim aktuellen Simulationsschritt
    double tempDifferenceAngle = 0;
    // Abweichung des tempPlayers
    double tempPlayerDifferenceAngle = 0;
    // Toleranz ( 10 Grad )
    double toleranceAngle = radFromDegree ( 10 );
    // Maximale Abweichung von Vorzugsrichtung in Rad
    double maxDeviationAngle;
    if ( globalWorldModel->role() == GOALIE_ROLE ) {
      maxDeviationAngle = radFromDegree( 60 );
    }
    else {
      maxDeviationAngle = radFromDegree( 40 );
    }
    // Schleifendurchlaufvariable
    double maxDeviationForLoop = ( ( maxDeviationAngle * 2 ) / 
        standardDifferenceAngle );
    // Winkel der Vorzugsrichtung
    double preferredAngle = fabs ( ( globalWorldModel->preferredKickDirection
        ( player.pos() ) ).angle() ) ;

    // Mich selbst aus scenario streichen
    scenario.removePlayerWithId ( globalWorldModel->id() );
    // Torwart raus. Das heisst wir nehmen an, 11 ist der Torwart.
    try {
      scenario.removePlayerWithId( 430 );
    }
    catch ( Situation::NoPlayerWithThisIdException ) {
      // Wenn er nicht drin ist, muss er auch nicht rausgeschmissen
      // weden
    }

    
    // Spieler die im Abseits stehen streichen
    INFO( Log::ACTIVE_OPTIONS, 5, "Abseitsueberpruefung ! \n");
 
    for ( int i = ( scenario.playerCount() -1 ); i >= 0; i--) {  
      if ( scenario.playerWithIndex(i).posReliability().get() > 
          Reliability::RANDOM ) {
        scenario.removePlayerWithIndex(i);
        continue;
      }
      else {
        if ( scenario.playerWithIndex(i).team() == OUR_TEAM_ID ) {
          if ( scenario.isOffsideWithIndex(i) ) {
            scenario.removePlayerWithIndex(i);
          } 
        }
      }   
    }

    INFO( Log::ACTIVE_OPTIONS, 5, "Abseitsueberpruefung beendet. \n");

    // Spieler , die ausserhalb des Einflussbereiches stehen aus
    // Situation streichen

    INFO( Log::ACTIVE_OPTIONS, 5, "Spieler aus Situation streichen ! \n");
 
    for ( int i = ( scenario.playerCount() -1); i >= 0; i--) {  
      tempPlayer = scenario.playerWithIndex(i);
      Vector toTarget ( tempPlayer.pos() - player.pos() );
      tempPlayerDifferenceAngle = fabs ( preferredAngle - 
          fabs ( toTarget.angle() ) );

      INFO( Log::ACTIVE_OPTIONS, 8, "tempPlayerDifferenceAngle: " + 
          String ( tempPlayerDifferenceAngle ) + " \n" );

      if ( tempPlayerDifferenceAngle > maxDeviationAngle + 
          toleranceAngle ) {
        scenario.removePlayerWithIndex(i); 
      }                              
    }
    INFO( Log::ACTIVE_OPTIONS, 5, "Situation bereinigt. \n"); 
    
    toTarget = globalWorldModel->preferredKickDirection( player.pos() );
    tempDifferenceAngle = 0;
    // bisher beste der Richtungen, wo der Gegner zuerst den Ball kriegt
    double bestBadAngle = toTarget.angle();
    // Die Richtung bestBadAngle hat die Bewertung bestBadAngleValue
    double bestBadAngleValue = 0;
    for ( int i = 0; i <= maxDeviationForLoop; i++) {
      /* Zielvarianz festlegen 
        Reihenfolge: Vorzugsrichtung +10, -10, ... , +40, -40 Grad */
      if ( i % 2 == 0 ) {
        tempDifferenceAngle += -standardDifferenceAngle * i;
      }
      else {
        tempDifferenceAngle += standardDifferenceAngle * i; 
      } 
      
      /* Zielberechnung */
      toTarget.setPolar( 1.0, toTarget.angle() + tempDifferenceAngle );

      /* Simulationsbeginn */
      INFO( Log::ACTIVE_OPTIONS, 5, "Simulation beginnt. \n");
      try {
        ball = globalWorldModel->ball();
        speed = globalConf->maxEffectiveBallSpeed * toTarget;
        ball.setSpeed( speed );
        ball.simulateSteps( 1 );
        scenario.setBall( ball );
        wasSimulationSuccessful = ( 
            scenario.stepsOfNearestPlayer( THEIR_TEAM_ID ) > 
            scenario.stepsOfNearestPlayer( OUR_TEAM_ID ) + 
            globalConf->passLeadForTeammates );
        if  ( wasSimulationSuccessful == YES ) {
          INFO( Log::ACTIVE_OPTIONS, 9, "Erfolgreiche Simulation!\n" );
          break;
        }
        else {
          double badAngleValue = scenario.stepsOfNearestPlayer( THEIR_TEAM_ID );
          INFO( Log::ACTIVE_OPTIONS, 9, "Gegner bekommt den Ball\n" );
          INFO( Log::ACTIVE_OPTIONS, 11, "badAngleValue = " +
              String( badAngleValue ) + "\n" );
          if ( bestBadAngleValue < badAngleValue ) {
            bestBadAngleValue = badAngleValue;
            bestBadAngle = tempDifferenceAngle;
            INFO( Log::ACTIVE_OPTIONS, 11, 
                "Ist bisher bester schlechter Winkel\n" );
          }
        }
      } /* try */
      catch ( Situation::NoPlayerWithThisIdException ) { 
        break; 
      }
      catch ( InvalidIdException ) { 
        break;
      }
    }
    if ( wasSimulationSuccessful ) {
      double angleFactor = 1 - fabs( tempDifferenceAngle) / maxDeviationAngle;
      double utilityRange = 
        globalConf->maxForwardPassUtility - globalConf->minForwardPassUtility;
      bestUtility = globalConf->minForwardPassUtility + 
          (int)( utilityRange * angleFactor );
      INFO( Log::ACTIVE_OPTIONS, 9, "angleFactor: " + String( angleFactor ) +
          "\nutilityRange: " + String( globalConf->minForwardPassUtility ) +
          " - " +  String( globalConf->maxForwardPassUtility ) +"\n" );
    }
    else {
      // 3 Moeglichkeiten fuer Fallback:
      // 1- Schuss auf Elfmeterpunkt im gegnerischen Strafraum
      // 2- Schuss in beste der schlechten Richtungen
      // 3- Schuss zur Feldmitte
      // -----
      // Stuermer und Mittelfeldspieler: Wenn im Abseits, 3, sonst 1
      // Andere (incl. Torwart): 1
      INFO( Log::ACTIVE_OPTIONS, 8, "Using fallback\n" );
      int role = globalWorldModel->role();
      if ( role == FORWARD_ROLE || role == LEARNING_FORWARD_ROLE ||
        role == MIDFIELD_ROLE || role == LEARNING_MIDFIELD_ROLE ) {
        if ( globalWorldModel->situation().offsideLineX() - 5.0 < 
            player.pos().x() ) {
          if ( globalWorldModel->region( player.pos() ) != 9 &&
              globalWorldModel->region( player.pos() ) != 10 ) {
            // Ein Stuermer/Mittelfeldspieler im Abseits, der nicht
            // in einer der vorderen Eckregionen steht, spielt zur Mitte
            INFO( Log::ACTIVE_OPTIONS, 8, "Stuermer im Abseits\n" );
            if ( player.pos().y() > 0 ) {
              toTarget.set( 1.0, -globalConf->ballSpeedMax );
            }
            else {
              toTarget.set( 1.0, globalConf->ballSpeedMax );
            }
          }
          else {
            // In den vorderen Eckregionen in Vorzugsrichtung schiessen
            toTarget = globalWorldModel->preferredKickDirection( player.pos() );
          }        
        }
        else {
          // nicht im Abseits, aber in gegnerischer Haelfte :
          // Schuss auf Elfmeterpunkt in den gegnerischen Strafraum
          INFO( Log::ACTIVE_OPTIONS, 8, "Schuss in gegnerischen Strafraum \n" );
          if ( player.pos().x() > 0 )
          toTarget = Vector( 41.0, 0.0) - player.pos();
        }
      } 
      else {
        // Verteidiger und Torwart hauen den Ball weg
        INFO( Log::ACTIVE_OPTIONS, 8, 
            "Schuss in beste der schlechten Richtungen\n" );
        // Verteidiger am Spielfeldrand spielt auf Punkt in gegnerischer
        // Spielhaelfte ( ebenfalls am Rand )
        if ( player.pos().x() <= ( -15 ) ) {
            if ( ( player.pos().y() >= 25 ) || ( player.pos().y() <= -25 ) ) {
              toTarget = Vector( 20.0, player.pos().y());
            }
        }
        else {
        // Verteidiger und Torwart hauen den Ball in die fuer den Gegner
        // schlechteste Richtung 
        toTarget.setPolar( 1.0, toTarget.angle() + bestBadAngle );
        toTarget = globalConf->ballSpeedMax * toTarget;
      }
        }
      
      bestUtility = globalConf->minForwardPassUtility;
    }
  }
  else {
    // nichts machen. Die default-Werte sind ausreichend.
  }
  INFO( Log::ACTIVE_OPTIONS, 5, "Utility " + String( bestUtility ) +
      " fuer Schuss in Richtung " + toTarget.toString() + "\n" );
  return bestUtility;
}


StorableObject* ForwardPassOption::copy() const 
{
  return new ForwardPassOption( *this );
}

String ForwardPassOption::toString() const  
{
  return String( "( ForwardPassOption utility: " ) +
    String( expectedUtility() ) + String( " ) " + toTarget.toString() );
}



 



