
/*
 * Aktueller Stand
 *
 *   $RCSfile: GoToBaseOption.cc,v $
 *   $Revision: 1.25 $
 *   $Date: 1998/08/14 13:20:36 $
 *   $Author: myritz $ 
 *
 */
 
#include "GoToBaseOption.h"
#include "Constraints.h"
#include "WorldModel.h"
#include "Skills.h"
#include "StraightLine.h"

GoToBaseOption::GoToBaseOption() : PassiveOption() 
{
  theId = GO_TO_BASE_OPTION_ID;
}  


int GoToBaseOption::baseUtility() const 
{
  INFO( Log::PASSIVE_OPTIONS, 2, "ENTER GoToBaseOption::baseUtility\n" );
  const ModelledBall& ball = globalWorldModel->ball();
  const ModelledPlayer& player = globalWorldModel->player();
  const Situation& situation = globalWorldModel->situation();

  int utility = PassiveOption::baseUtility();
  setIsFreeKickTargetSelected( NO );
  setIsMoveForwardTargetSelected( NO );
  double distFromBall = ( ball.pos() - player.pos() ).length();
  distFromBall -= ball.size() + player.size();

  // Zielposition in Abh"angigkeit von Spielzustand setzen
  setTarget( globalWorldModel->basePosition() );

  if( globalWorldModel->ball().posReliability().get() < 
      Reliability::VERY_GOOD ) {
    if(  isDefenseSituation() ) {
      resetDefenderTarget();
    }
    resetTargetIfMoveForwardSituation();
  }

  setTarget( closestNonOffsideTarget( target() ) );

  if( utility == NEUTRAL_UTILITY ) { // keine Daten "uber mich
    if( globalWorldModel->gameState() == WorldModel::BEFORE_KICK_OFF ||
        globalWorldModel->gameState() == WorldModel::GOAL ) {
      // ich weiss nicht wo ich bin, kann aber beamen
      utility = MAX_UTILITY;
    }
  }
  else if( utility == MAX_UTILITY ) { // Daten ueber mich, und weg vom Ball
    utility = NEUTRAL_UTILITY;

    // wenn wir vorruecken koennen wir ungenauer werden 
    double gotoBasePositionAccuracy = globalConf->gotoBasePositionAccuracy;
    if( isMoveForwardTargetSelected() ) {
      gotoBasePositionAccuracy = 3 * globalConf->gotoBasePositionAccuracy;
    }
    if( player.distTo( target() ) < gotoBasePositionAccuracy ) {
      // schon am Zielpunkt
      utility = NEUTRAL_UTILITY;
    }
    else {
      // noch nicht am Ziel
      switch ( globalWorldModel->gameState() ) {
        case WorldModel::BEFORE_KICK_OFF:
        case WorldModel::GOAL:
          utility = MAX_UTILITY;
          break;
        /*
        case WorldModel::PLAY_ON:

          if ( distFromBall < globalConf->minUtilitydist ) {
            utility = MIN_UTILITY;
          }
          else {
            if ( distFromBall > globalConf->maxUtilitydist ) {
              utility = MAX_UTILITY;
            }
            else {
              // normieren auf MAX und MIN_UTILITY
              utility = MIN_UTILITY + 
                  ( int ) ( ( double ) ( MAX_UTILITY - MIN_UTILITY ) / 
                  ( double ) ( globalConf->maxUtilitydist -
                  globalConf->minUtilitydist) *
                  ( distFromBall - globalConf->minUtilitydist ) );
            }
          }
          break;
        */

        default:
          if( situation.offsideLineXReliability().get() < Reliability::BAD ) {
            utility = globalConf->normalGoToBaseUtility;
          }
          else {
            utility = globalConf->minGoToBaseUtility;
          }
          if( globalWorldModel->decisionReason() == 
              WorldModel::DECISION_REASON_SEE_INFO &&
              ball.posReliability().get() <= Reliability::VERY_GOOD ) {
            setForceContinuationTime( 
                globalWorldModel->situation().stepsOfNearestPlayer() );
        }
      }
    }
  }
  else { // Daten "uber mich, und am Ball
    // doNothing
  }  
  
  // Hier wird im rasteredPlayground fuer "lernende" Agenten 
  // waehrend PLAY_ON erfasst, wo sich Ball und Gegner aufhalten.
  if ( globalWorldModel->gameState() == WorldModel::PLAY_ON ) {
    switch ( globalWorldModel->role() ) {
      case LEARNING_FORWARD_ROLE :
      case LEARNING_MIDFIELD_ROLE :
      case LEARNING_DEFENDER_ROLE :
        if ( distFromBall < ( double ) globalConf->ballLearnDist ){
          if ( ball.posReliability().get() <= Reliability::GOOD ) {
            globalWorldModel->rasteredPlayground().
                acquireKnowledgeAbout( BALL_ID );
            INFO( Log::PASSIVE_OPTIONS, 12, "\nBall gelernt" );
          }
        }
        else {
          INFO( Log::PASSIVE_OPTIONS, 12,"Lerne Player\n" );
              globalWorldModel->rasteredPlayground().
                  acquireKnowledgeAbout( PLAYER_ID );
          INFO( Log::PASSIVE_OPTIONS, 12,"\nPlayer gelernt\n" );
        }
      default: 
        break;
    } 
  } 

  INFO( Log::PASSIVE_OPTIONS, 6, "utility: " + String( utility ) + "\n" );
  INFO( Log::PASSIVE_OPTIONS, 2, "LEAVE GoToBaseOption::baseUtility\n" );
  
  return utility;
}


bool GoToBaseOption::isPossible() const 
{
  if ( isOptionPossible() == YES ) {
    setTarget( globalWorldModel->rasteredPlayground().goodPassivePosition() );
    if( isFreeKickTargetSelected() ) {
      resetDefenderTarget();
    }
    else if( isMoveForwardTargetSelected() ) {
      resetTargetIfMoveForwardSituation();
    }
    // immer Abseits beachten
    setTarget( closestNonOffsideTarget( target() ) );
    double distFromTarget =
      globalWorldModel->player().distTo( target() );
    if ( globalWorldModel->player().posReliability().get() < 
         Reliability::BAD &&
         distFromTarget <= globalConf->gotoBasePositionAccuracy ) {
      setIsOptionPossible( NO );
    }
  }
  return isOptionPossible();
}



Plan GoToBaseOption::generatePlan() const 
{
  Plan plan;
  if( !isTargetComputed() ) {
    throwException( Exception( "Target should had been already computed" ) );
  }
  double staminaLimit = globalConf->gotoBaseStaminaLimit;
  if( isConstraintAccepted( AVOID_OFFSIDE_CONSTRAINT_ID ) ) {
    staminaLimit -= globalConf->avoidOffsideStaminaLimitBonus;
  }
  if( isMoveForwardTargetSelected() ) {
    staminaLimit += 200;
  }
  
  globalSkills->resetSituation();
  plan.addPlan( globalSkills->setViewWidth( VIEW_WIDTH_WIDE ) );
  plan.addPlan( isSomethingToSay() );
  
  // doppelter Code - Siehe Zeile 61
  double gotoBasePositionAccuracy = globalConf->gotoBasePositionAccuracy;
  if( isMoveForwardTargetSelected() ) {
    gotoBasePositionAccuracy = 3 * globalConf->gotoBasePositionAccuracy;
  }
  plan.addPlan( movementTo( target(), gotoBasePositionAccuracy,
                            staminaLimit ) );
  if ( plan.length() == 0 ) {
    setIsPlanCompleted( YES );
  }
  return plan;
}

void GoToBaseOption::resetTargetIfMoveForwardSituation() const {
  if( ( globalWorldModel->gameState() != WorldModel::BEFORE_KICK_OFF &&
	    globalWorldModel->gameState() != WorldModel::GOAL ) &&
      ( globalWorldModel->role() == MOVE_FORWARD_DEFENDER_ROLE || 
	    globalWorldModel->role() == MOVE_FORWARD_MIDFIELD_ROLE ) ) {
    double moveForwardBallDistActivation;
    double moveForwardStopLine;
    if( globalWorldModel->role() == MOVE_FORWARD_DEFENDER_ROLE ) {
      moveForwardBallDistActivation = 
	        globalConf->defenderMoveForwardBallDistActivation;
      moveForwardStopLine = globalConf->defenderMoveForwardStopLine;
    }
    else if( globalWorldModel->role() == MOVE_FORWARD_MIDFIELD_ROLE ) {
      moveForwardBallDistActivation = 
	        globalConf->midfieldMoveForwardBallDistActivation;
      moveForwardStopLine = globalConf->midfieldMoveForwardStopLine;
    }

    const ModelledBall& ball = globalWorldModel->ball();
    if( ball.pos().x() - moveForwardStopLine >
	     moveForwardBallDistActivation ) {
      double bestXPos = target().x() + ( ball.pos().x() - 
	        moveForwardStopLine - moveForwardBallDistActivation );
      bestXPos = min( bestXPos, moveForwardStopLine );

      if( globalWorldModel->role() == MOVE_FORWARD_DEFENDER_ROLE ) {
	      double distFromSideLine = Field::fieldWidth() / 2 - 
	          fabs( target().y() );
	      double bestYPos = target().y();
        if( distFromSideLine >= 7.0 ) {
	        double maxShift = distFromSideLine - 7.0;
	        double shiftFactor = 0.5 * fabs( target().y() ) / distFromSideLine;
	        maxShift = min( maxShift, shiftFactor * fabs( target().y() ) );
	        shiftFactor = ( bestXPos - target().x() ) / ( 45.0 - 
                    ( moveForwardStopLine + moveForwardBallDistActivation ) );
	        bestYPos += maxShift * shiftFactor * signum( target().y() );
        }
	      setTarget( Vector( bestXPos, bestYPos ) );
      }
      else if( globalWorldModel->role() == MOVE_FORWARD_MIDFIELD_ROLE ) {
	      setTarget( Vector( bestXPos, target().y() ) );
      }

      setIsMoveForwardTargetSelected( YES );
    }
  }
}

void  GoToBaseOption::resetDefenderTarget() const
{
  INFO( Log::PASSIVE_OPTIONS, 8, 
        "ENTER GoToBaseOption::resetDefenderTarget\n" );

  if( globalWorldModel->role() == DEFENDER_ROLE || 
      globalWorldModel->role() == LEARNING_DEFENDER_ROLE ||
      globalWorldModel->role() == MOVE_FORWARD_DEFENDER_ROLE ) {
    const ModelledBall& ball = globalWorldModel->ball();
    if( ball.pos().x() -5 < target().x() ) {
      setTarget( Vector( max( ball.pos().x() - 5, -45 ), target().y() ) );
    }
    double distHomeToBall = ( target() - ball.pos() ).length();
    
    StraightLine ballToPlayerHome( ball.pos(), target() );
    Vector bestPosition =
      ballToPlayerHome.calcPoint( globalConf->defenderFreeKickDist );
    
    if( distHomeToBall > globalConf->defenderFreeKickDist ) {
      double distHomeToBestPos = ( bestPosition - target() ).length();
      distHomeToBestPos = min( distHomeToBestPos, 
				   globalConf->defenderFreeKickMoveUp );
      setTarget( ballToPlayerHome.calcPoint( distHomeToBall - 
					     distHomeToBestPos ) );
    }
    else {
      setTarget( bestPosition );
    }
    setIsFreeKickTargetSelected( YES );
  }

  if( globalWorldModel->role() == MIDFIELD_ROLE || 
      globalWorldModel->role() == LEARNING_MIDFIELD_ROLE ||
      globalWorldModel->role() == MOVE_FORWARD_MIDFIELD_ROLE ) {
    const ModelledBall& ball = globalWorldModel->ball();
    if( ball.pos().x() -10 < target().x() ) {
      setTarget( Vector( max( ball.pos().x() - 10, -25 ), target().y() ) );
    }
    setIsFreeKickTargetSelected( YES );
  }

  INFO( Log::PASSIVE_OPTIONS, 8,
	"reset Target to: " + target().toString() + "\n" );
  INFO( Log::PASSIVE_OPTIONS, 8, 
        "LEAVE GoToBaseOption::resetDefenderTarget\n" );
}

StorableObject* GoToBaseOption::copy() const 
{
  return new GoToBaseOption( *this );
}

String GoToBaseOption::toString() const  
{
  return String( "( GoToBaseOption" ) + PassiveOption::toString() +
    String( " ) " );
}


 

