#ifndef _GOAL_KICK_OPTION_H_
#define _GOAL_KICK_OPTION_H_

#include "KickOption.h"
#include "Field.h"

const int GOAL_KICK_OPTION_ID = 703;


/**
  Diese Klasse steht fuer den Wunsch den Ball in das gegnerische Tor zu
  schiessen. Dieser Wunsch
  bekommt eine hohe Aktivierung, wenn ein Torschuss aussichtsreich erscheint.
  Dieser Wunsch generiert auf Anfrage
  die gleichnamige Intention, die fuer diesen Wunsch dann einen optimalen
  Balltritt ermittelt.

  {\small {\bf Quelltext}

  [\URL[Header]{src/GoalKickOption.h} | [\URL[Implementierung]{src/GoalKickOption.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: GoalKickOption.h,v $
  *   $Revision: 1.10 $
  *   $Date: 1998/08/14 13:20:39 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class GoalKickOption : public KickOption
{
public:
  /** Erzeugt eine GoalKickOption. */
  GoalKickOption();

  /** Beschreibung als String fuer die Ausgabe auf einem Terminal. */
  virtual String toString() const;

  /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. Diese Methode muss fuer JEDE Unterklasse neu definiert 
      werden. */
  virtual StorableObject* copy() const;

  /** Spezieller Plan, der noch eine Korrektur enthaelt, damit die 
    fuer die Koerperbreite korrigiert wird. */
  virtual Plan generatePlan() const;
  
protected:
  
  /** Der Grundnutzen des Wunsches. Dieser Teil der Nutzenberechnung ist nur 
      vom Weltmodell abhaengig.
      @return Im Allgemeinen ein Wert zwischen MAX_UTILITY und MIN_UTILITY. */
  int baseUtility() const;

  /** Passt den Zielvektor so an, dass
    die Koerperbreite kompensiert wird. */
  Vector adjustedToTarget( const Vector& toTarget ) const;

};

#endif



