/*
 * Aktueller Stand
 *
 *   $RCSfile: GoalieOption.cc,v $
 *   $Revision: 1.11 $
 *   $Date: 1998/08/14 13:20:40 $
 *   $Author: myritz $ 
 *
 */
 
#include "GoalieOption.h"
#include "Constraints.h"
#include "WorldModel.h"
#include "Skills.h"

int GoalieOption::theFreeKickSimStep = 0;

GoalieOption::GoalieOption() : PassiveOption()
{
  goalieDistFromGoal = globalConf->goalieDistFromGoal;
  isBetterAsGoalie = 5;
  setAllFlags();
  theId = GOALIE_OPTION_ID;
}  


int GoalieOption::baseUtility() const 
{
  INFO( Log::PASSIVE_OPTIONS, 2, "ENTER GoalieOption::baseUtility\n" );

  int utility;
  const ModelledBall& ball = globalWorldModel->ball();
  const ModelledPlayer& player = globalWorldModel->player();
  setForceContinuationTime( NO_CONTINUATION );
  

  if ( globalWorldModel->role() != GOALIE_ROLE ) {
    INFO( Log::PASSIVE_OPTIONS, 10, "Ich bin kein Torwart\n" );
    utility = MIN_UTILITY;
  }
  else if ( ball.posReliability().get() >= Reliability::VERY_GOOD ) {
    INFO( Log::PASSIVE_OPTIONS, 10, "Keine ausreichende Zuverlaessigkeit\n" );
    utility = MIN_UTILITY;
  }
  else if( globalWorldModel->gameState() == WorldModel::FREE_KICK &&
      globalWorldModel->initiative() == OUR_TEAM_ID ) {
    INFO( Log::PASSIVE_OPTIONS, 12, "Ein Mitspieler wird den Freistoss schiessen\n");
    setFreeKickSimStep( globalWorldModel->simStep() );
    setGoToGoalieHomePosition( YES );
    utility = MAX_UTILITY;
  } 
  else if ( globalWorldModel->gameState() == WorldModel::GOAL ) {
    INFO( Log::PASSIVE_OPTIONS, 10, "Alles vorbei. jemand hat ein Tor geschossen.\n" );
    setGoToGoalieHomePosition( YES );
    utility = MAX_UTILITY;  
  }
  else if ( isDangerousPos( ball.pos() ) ) {
    INFO( Log::PASSIVE_OPTIONS, 10, "Ball kommt gefaehrlich nahe.\n" );
    if ( isCatchPossible() ) {
      INFO( Log::PASSIVE_OPTIONS, 12, "Torwart versucht den Ball zu erreichen.\n");
      setCatchAtOnce( YES );
      utility = MAX_UTILITY;
    }
    else if ( globalWorldModel->situation().nearestPlayer( 
        BOTH_TEAMS_ID ).id() == globalWorldModel->id() && 
        !( globalWorldModel->gameState() == WorldModel::FREE_KICK &&
        globalWorldModel->initiative() == THEIR_TEAM_ID ) ) {
      INFO( Log::PASSIVE_OPTIONS, 12, "Torwart erreicht ball als Erster.\n");
      if ( fabs( ball.pos().y() ) < 
          globalWorldModel->field().penaltyAreaWidth() / 2.0 ) {
        // InterceptBall soll anspringen
        utility = MIN_UTILITY;
      }
      else {
        // Wir wollen nicht so weit vom Tor weg. Lieber doch Positionsspiel
        setPositionPlay( YES );
        utility = MAX_UTILITY;
      }            
    }
    else if ( isBallInPenaltyArea( ball.pos() ) ) {
      INFO( Log::PASSIVE_OPTIONS, 10, "Ball ist im Strafraum.\n" );
      setPositionPlay( YES );
      utility = MAX_UTILITY;
    }
    else {
      utility = MAX_UTILITY;
      if ( player.distTo( globalWorldModel->ourGoal() )  < 20 ) {
        INFO( Log::PASSIVE_OPTIONS, 10, "Ich stehe nah am Tor.\n" );
        setPositionPlay( YES );
      }
      else {
        INFO( Log::PASSIVE_OPTIONS, 10, "Ich bin weit weg vom Tor.\n" );
        setGoToGoalieHomePosition( YES );
      }
    }
  }
  else {
    INFO( Log::PASSIVE_OPTIONS, 10, "Ball ist weit weg. Gehe auf Heimatposition.\n" );
    setGoToGoalieHomePosition( YES );
    utility = MAX_UTILITY;
  }
             
  INFO( Log::PASSIVE_OPTIONS, 6, "utility: " + String( utility ) + "\n" );
  INFO( Log::PASSIVE_OPTIONS, 2, "LEAVE GoalieOption::baseUtility\n" );

  setIsOptionPossible( utility > 0 );  
  return utility;
}


int GoalieOption::constraintCost( const Constraint *constraint ) const 
{
  int cost;
  switch( constraint->id() ) {
    case CONSERVE_STAMINA_CONSTRAINT_ID:
      if ( globalWorldModel->gameState() == WorldModel::BEFORE_KICK_OFF ||
           globalWorldModel->gameState() == WorldModel::GOAL ) {
        cost = NEUTRAL_COST;
      }
      else {
        if ( isCatchAtOnce() ) {
          cost = MIN_COST;
        }
        else if ( isGoalieIntercept() ) {
          cost = MAX_COST;
        }
        else {
          cost = NEUTRAL_COST;
        }
      }
      break;
    case UNDEF_CONSTRAINT_ID:
      cost = NEUTRAL_COST;
      break;
    default:
      cost = MAX_COST;
  }
  INFO( Log::PASSIVE_OPTIONS, 7, "Constraint cost: " + String( cost ) +"\n" );
  return cost;
}


bool GoalieOption::forceContinuation() const
{
  INFO( Log::PASSIVE_OPTIONS, 6,
        "ENTER GoalieOption::forceContinuation\n");
  bool isToContinue;

  if ( globalWorldModel->decisionReason() == 
      WorldModel::DECISION_REASON_REFEREE_DECISION ) {
    INFO( Log::PASSIVE_OPTIONS, 11, "Schiedrichterentscheidung : Nein.\n" );
    isToContinue = NO;
  }
  else {
    if ( globalWorldModel->ball().posReliability().get() < 
        Reliability::VERY_GOOD ) {
      INFO( Log::PASSIVE_OPTIONS, 11, "Zuverlaessigkeit ist zu schlecht: Nein.\n" );
      isToContinue = NO;
    }
    else {
      if ( forceContinuationTime() >= globalWorldModel->simStep() ) {
        INFO( Log::PASSIVE_OPTIONS, 11, "Continuation time ist gut : Ja.\n" );
        isToContinue = YES;
      }
      else {
        INFO( Log::PASSIVE_OPTIONS, 11, "Continuation time ist schlecht: Nein\n" );
        isToContinue = NO;
      }
    }
  }
  INFO( Log::PASSIVE_OPTIONS, 5,
        String( "forceContinuation von der GoalieOption: " ) + 
        ( ( isToContinue ) ? "JA" : "NEIN" ) + "\n" );
  return isToContinue;
}

Plan GoalieOption::movementTo(
      const Vector& target, 
      const double targetRadius, 
      const double staminaLimit,
      const int specialBehaviors ) const
{
  const ModelledPlayer& player = globalWorldModel->player();
  const ModelledBall& ball = globalWorldModel->ball();
  Plan plan;
  
  if ( player.distTo( target ) >= targetRadius ) {
    // Zur Position hinkommen
    plan = PassiveOption::movementTo( target, targetRadius, staminaLimit,
        specialBehaviors );
  }
  else {
    double relativeBallDir;
    relativeBallDir = normalizeAngle( ( ball.pos() - player.pos() ).angle() );
    if ( isAlmostEqual( normalizeAngle( player.faceDir() ), 
        relativeBallDir ) == NO ) {
      // Wenn wir schon da sind, direkt zum Ball schauen
      plan = globalSkills->turnTo( relativeBallDir );
    }
    else {
      // Es gibt wirklich nichts zu tun.
      plan = globalSkills->doNothing();
    }
  }
  return plan;
}


Plan GoalieOption::generatePlan() const
{
  INFO( Log::PASSIVE_OPTIONS, 6,
        "ENTER GoalieOption::generatePlan\n");
  Plan plan;
  const ModelledBall& ball = globalWorldModel->ball();
  const ModelledPlayer& player = globalWorldModel->player();

  globalSkills->resetSituation();
  double distBefore = player.distTo( ball.pos() );
  if ( distBefore > 6 ) {
    // Wenn sich der Ball schnell auf den Torwart zubewegt,
    // auf NARROW schalten
    ModelledBall simBall( ball );
    simBall.simulateSteps();
    double distAfter = player.distTo( ball.pos() );
    if ( distBefore - distAfter > 1.5 || 
        distBefore < 6.0 ) {
      plan.addPlan( globalSkills->setViewWidth( VIEW_WIDTH_NARROW ) );
    }
    else {
      plan.addPlan( globalSkills->setViewWidth( VIEW_WIDTH_NORMAL ) );
    }
  }

  if( isCatchAtOnce() ) {
    INFO( Log::PASSIVE_OPTIONS, 11, "Versuche Catch-Kommando." );
    double catchAngle;
    if ( ball.speed().length() > 1 ) {
      ModelledBall simBall( ball );
      simBall.simulateSteps();
      catchAngle = player.angleTo( simBall.pos() );
    }
    else {
      catchAngle = player.angleTo( ball.pos() );
    }
    plan.addAction( globalSkills->simpleCatch( catchAngle ) );
  }
  else {
    if ( isGoToGoalieHomePosition() ) {
      computeHomePosition();
      plan.addPlan( movementTo( target(), 1.5, 
          globalConf->gotoBaseStaminaLimit ) );
    } 
    else if ( isPositionPlay() ) {
      computePositionPlay();
      if ( isBackwardsRunUseful() ) {
        plan.addPlan( movementTo( target(), 1.5, 
            globalConf->getIntoSpaceStaminaLimit, Skills::RUN_BACKWARDS ) );
      }
      else {
        plan.addPlan( movementTo( target(), 1.5, 
            globalConf->getIntoSpaceStaminaLimit, Skills::AVOID_OBSTACLES ) );
      }
    }
    INFO( Log::PASSIVE_OPTIONS, 10, "target: " + target().toString() + "\n" );
  }
  INFO( Log::PASSIVE_OPTIONS, 10, "plan: " + plan.toString() + "\n" ); 
  INFO( Log::PASSIVE_OPTIONS, 6, "LEAVE GoalieOption::generatePlan\n");
  return plan;
}

bool GoalieOption::isBackwardsRunUseful() const
{
  const ModelledPlayer& player = globalWorldModel->player();
  Vector fromTarget = player.pos() - target();
  Vector toBall =  globalWorldModel->ball().pos() - player.pos();
  double angle = fabs( fromTarget.angle() - toBall.angle() );
  return ( player.visibleAngle() > angle );
}

bool GoalieOption::isTeamMateInBallZone() const
{
  const Situation& situation = globalWorldModel->situation();
  const ModelledPlayer& player = globalWorldModel->player();
  ModelledPlayer teammate;
  const ModelledBall& ball = globalWorldModel->ball();
  StraightLine playerToBall( player.pos(), ball.pos() );
  
  double distToBall = player.distTo( ball.pos() );
  double distToTarget = player.distTo( target() );
  
  for ( int index = 0; index < situation.playerCount(); index++ ) {
    teammate = situation.playerWithIndex( index );
    if ( teammate.team() == OUR_TEAM_ID ) {
      if ( isDangerousPos( teammate.pos() ) &&
          teammate.id() != player.id() ) {
        double distOfTeammateToLine = 
          playerToBall.distTo( teammate.pos() );
        if ( distOfTeammateToLine <= distToTarget ) {
          double distToTeammate = player.distTo( teammate.pos() );
          if ( distToTeammate <= distToBall ) {
            return YES;
          }
        }
      }
    }
  }
  return NO;
}
	  
StorableObject* GoalieOption::copy() const 
{
  return new GoalieOption( *this );
}


String GoalieOption::toString() const  
{
  return String( "( GoalieOption" ) + PassiveOption::toString() +
    String( " ) " );
}


bool GoalieOption::isDangerousPos( const Vector& aPos ) const
{
  if ( ( globalWorldModel->ourGoal().x() + 
      globalConf->distToBallForEvaluation ) > aPos.x() ) {
    return YES;
  }
  else {
    return NO;
  }
}


bool GoalieOption::isBallInPenaltyArea(const Vector& aVector) const
{
  if( aVector.x() <=
      globalWorldModel->field().posOfPoint(UPPER_LEFT_PENALTY_FLAG_ID).x() )
    if( ( aVector.y() <=
	  globalWorldModel->field().posOfPoint(UPPER_LEFT_PENALTY_FLAG_ID).y() ) &&
	( aVector.y() >=
	  globalWorldModel->field().posOfPoint(LOWER_LEFT_PENALTY_FLAG_ID).y() ) )
      return YES;
  return NO;
}

  
void GoalieOption::computeHomePosition() const
{
  INFO( Log::PASSIVE_OPTIONS, 6,
        "ENTER GoalieOption::computeHomePos\n");
  Vector goalieHomePos;
  if ( globalWorldModel->simStep() - freeKickSimStep() > 6 ) {
    goalieHomePos.set( globalWorldModel->ourGoal().x() + 
      goalieDistFromGoal, globalWorldModel->ourGoal().y() );
  }
  else {
    goalieHomePos.set( globalWorldModel->ourGoal().x() + 2, 
        globalWorldModel->ourGoal().y() );
  }
  setTarget( goalieHomePos );
  INFO( Log::PASSIVE_OPTIONS, 6, "LEAVE GoalieOption::computeHomePos\n");
}

bool GoalieOption::isCatchPossible() const
{
  if ( globalWorldModel->initiative() == BOTH_TEAMS_ID ||
      globalWorldModel->initiative() == OUR_TEAM_ID ) {
    if ( globalWorldModel->simStep() - globalWorldModel->simStepOfLastCatch() >
        globalConf->catchCycle && 
        globalWorldModel->simStep() - freeKickSimStep() > 6 ) {
      if ( isBallInPenaltyArea( globalWorldModel->ball().pos() ) && 
          globalWorldModel->player().distTo( 
          globalWorldModel->ball().pos() ) 
          < globalConf->catchAreaLength ) {
        return YES;
      }
    }
  }
  INFO( Log::PASSIVE_OPTIONS, 5, "Catch is not possible.\n");
  INFO( Log::PASSIVE_OPTIONS, 8, "simStepOfLastCatch: " + 
      String( globalWorldModel->simStepOfLastCatch() ) + "\n" );
  INFO( Log::PASSIVE_OPTIONS, 8, "freeKickSimStep: " + 
      String( freeKickSimStep() ) + "\n" );
  return NO;
}


void GoalieOption::computePositionPlay() const
{
  INFO( Log::PASSIVE_OPTIONS, 6, "ENTER GoalieOption::computePositionPlay\n");
  if ( isBallInPenaltyArea( globalWorldModel->ball().pos() ) ){
    INFO( Log::PASSIVE_OPTIONS, 12, "Ball somewhwere in Penalty-Area\n");
    positionPlay();
  }
  else {
    INFO( Log::PASSIVE_OPTIONS, 12, "Outside of Penalty-Area\n");
    double addition = globalWorldModel->player().distTo( 
      globalWorldModel->ball().pos() );
    if ( addition <= 11 ) {
      positionPlay( 0 );
    }
    else {
      positionPlay( ( addition - 11 ) / 2.5 );
    }
  }
  INFO( Log::PASSIVE_OPTIONS, 6, "LEAVE GoalieOption::computePositionPlay\n");
}

void GoalieOption::positionPlay( double addition ) const
{ 
  INFO( Log::PASSIVE_OPTIONS, 6,
        "ENTER GoalieOption::positionPlay\n");

  // beste Position ist Winkelhalbierende des Winkels
  // TorpfostenTop-Ball-TorpfostenBottom in einer Entfernung von
  // goaliedistFromGoalius vom Tormitelpunkt entfernt.
  // Zusatzlich gehen wir dem Ball um addition entgegen.
  const ModelledBall& ball = globalWorldModel->ball();
  const ModelledPlayer& player = globalWorldModel->player();
      
  // Verhaeltnissatz: a/b == c/d ; goalWidth = c + d
  double a, b, c;
  // Sicherstellen, dass a und b nicht 0 sind.
  a = max( ( ball.pos() - globalWorldModel->ourGoalTop() ).length(), 0.1 );
  b = max( ( ball.pos() - globalWorldModel->ourGoalBottom() ).length(), 0.1 );
  c = globalConf->goalWidth / ( 1 + a / b );
  Vector direction( 0, 1 );
  
  Vector basePointInGoal = globalWorldModel->ourGoalBottom() + c * direction;
  StraightLine basePointToBall( basePointInGoal, ball.pos() );
  if ( ( player.pos() - basePointToBall.calcPoint( 
      goalieDistFromGoal + addition )).length() < 1 ) {
    // Weniger als 1m laufen wir nicht.
    setTarget( player.pos() );
  }
  else {
    setTarget( basePointToBall.calcPoint( goalieDistFromGoal + addition ) );
    if( isTeamMateInBallZone() ) {
      setTarget( basePointToBall.calcPoint( goalieDistFromGoal ) );
    }
  }
  INFO( Log::PASSIVE_OPTIONS, 6,
      "LEAVE GoalieOption::positionPlay\n");  
}
    
