
/**
 * Aktueller Stand
 *
 *   $RCSfile: ModelledPlayer.cc,v $
 *   $Revision: 1.48 $
 *   $Date: 1998/08/14 13:20:54 $
 *   $Author: myritz $ 
 *
 */

#include "ModelledPlayer.h"
#include "ModelledBall.h"
#include "Log.h"
#include "Skills.h"
#include "WorldModel.h"
#include <math.h>

double ModelledPlayer::thePlayerSize;
double ModelledPlayer::thePlayerMaxSpeed;
double ModelledPlayer::thePlayerDecay;
double ModelledPlayer::theStaminaIncMax;
double ModelledPlayer::theStaminaMax;
double ModelledPlayer::theRecoverDecThr;
double ModelledPlayer::theRecoverDec;
double ModelledPlayer::theRecoverMin;
double ModelledPlayer::theEffortDecThr;
double ModelledPlayer::theEffortDec;
double ModelledPlayer::theEffortIncThr;
double ModelledPlayer::theEffortInc;
double ModelledPlayer::theEffortMin;
double ModelledPlayer::theDashPowerRate;
double ModelledPlayer::theMaxPower;
double ModelledPlayer::theMinPower;
double ModelledPlayer::theMaxMoment;
double ModelledPlayer::theMinMoment;
double ModelledPlayer::theNormVisibleAngle;
double ModelledPlayer::theInertiaMomentOfTurn;
short  ModelledPlayer::isSetOnce=NO;

const int ModelledPlayer::MAX_STEPS_TO_BALL = 100;

ModelledPlayer::ModelledPlayer( int id,
  const Vector& aPos, 
  const Vector& aSpeed,
  double aFaceDir,
  const Reliability& aPosReliability,
  const Reliability& aSpeedReliability,
  const Reliability& aFaceDirReliability,
  int aViewWidth,
  double stamina,
  double effort,
  double recovery,
  const Reliability& aStaminaReliability ) 
  : ModelledObject( id, aPos, aSpeed, aPosReliability, aSpeedReliability),
    theFaceDir( aFaceDir ),
    theViewWidth( aViewWidth ),
    theStaminaReliability( aStaminaReliability ),
    theFaceDirReliability( aFaceDirReliability )
{
  if (ModelledPlayer::isSetOnce==NO){
    setClassVars();
  }
  setVisibleAngle( aViewWidth );
  setStamina( stamina );
  setEffort( effort );
  setRecovery( recovery );
}

ModelledPlayer::ModelledPlayer( const MovingPlayer& aMovingPlayer,  
  const Reliability& aPosReliability,
  const Reliability& aSpeedReliability,
  const Reliability& aFaceDirReliability,
  int aViewWidth,
  double stamina,
  double effort,
  double recovery,
  const Reliability& aStaminaReliability ) 
  : ModelledObject( aMovingPlayer.id(), aMovingPlayer.pos(), 
                    aMovingPlayer.speed(), aPosReliability, 
                    aSpeedReliability ),
    theFaceDir( aMovingPlayer.faceDir() ),
    theViewWidth( aViewWidth ),
    theStaminaReliability( aStaminaReliability ),
    theFaceDirReliability( aFaceDirReliability )
{
  if (isSetOnce==NO){
    setClassVars();
  }
  setVisibleAngle( aViewWidth );
  setStamina( stamina );
  setEffort( effort );
  setRecovery( recovery );
}

void ModelledPlayer::setClassVars() {
  setSize(globalConf->playerSize);
  setMaxSpeed(globalConf->playerSpeedMax);
  setDecay(globalConf->playerDecay);
  setStaminaIncMax(globalConf->playerStaminaInc);
  setStaminaMax(globalConf->playerStaminaMax);
  setRecoverDecThr(globalConf->playerRecoverDecThr);
  setRecoverDec(globalConf->playerRecoverDec);
  setRecoverMin(globalConf->playerRecoverMin);
  setEffortDecThr(globalConf->playerEffortDecThr);
  setEffortDec(globalConf->playerEffortDec);
  setEffortIncThr(globalConf->playerEffortIncThr);
  setEffortInc(globalConf->playerEffortInc);
  setEffortMin(globalConf->playerEffortMin);
  setDashPowerRate(globalConf->dashPowerRate);
  setMaxPower(globalConf->maxPower);
  setMinPower(globalConf->minPower);
  setMaxMoment(radFromDegree(globalConf->maxMoment));
  setMinMoment(radFromDegree(globalConf->minMoment));
  setNormVisibleAngle(globalConf->visibleAngle);
  setInertiaMomentOfTurn(globalConf->inertiaMoment);
  ModelledPlayer::isSetOnce=YES;
}

void ModelledPlayer::setStaminaValues(double newStamina = UNDEF_DOUBLE,
    double newEffortFactor = UNDEF_DOUBLE,
    double newRecoveryFactor = UNDEF_DOUBLE,
    const Reliability& aStaminaReliability )
{
  setStamina( newStamina );
  setEffort( newEffortFactor );
  setRecovery( newRecoveryFactor );
  setStaminaReliability( aStaminaReliability );
}

void ModelledPlayer::setViewWidth( int newViewWidth )
{
  theViewWidth = newViewWidth;
  setVisibleAngle( newViewWidth );
}

void ModelledPlayer::setVisibleAngle( int viewWidth ) 
{
  switch( viewWidth ) {
  case VIEW_WIDTH_NARROW:
    theVisibleAngle = normVisibleAngle() / 2;
    break;
  case VIEW_WIDTH_NORMAL:
    theVisibleAngle = normVisibleAngle();
    break;
  case VIEW_WIDTH_WIDE:
    theVisibleAngle = normVisibleAngle() * 2;
    break;
  default:
    throwException( ModelledPlayerException( "viewWidth undef" ) );
  }
}

void ModelledPlayer::simulateSteps( unsigned int simSteps = 1 ) 
{
  if( simSteps > 10 ) {
    WARNING( Log::MODELLED_OBJECTS, 
        "ModelledPlayer::simulateSteps: more than 10 steps." );
  }
  if( posReliability().get() < Reliability::RANDOM &&
      speedReliability().get() < Reliability::RANDOM ) {
    while( simSteps > 0 ) {    
      setPos( pos() + speed() );
      setSpeed( decay() * speed() );
    
    reduceReliability();
    theFaceDirReliability.decrease();
    recover( 1 );
    simSteps--;
    }
  }
}

void ModelledPlayer::accelerate(double power) 
{
  Vector myAcceleration;

  if( speedReliability().get() < Reliability::RANDOM ) {
    power = normalizeDashPower(power);
    if ( fabs(power) > theStamina ){
      power = signum(power) * theStamina ;
    }
    theStamina -= fabs(power);
    power = effectiveDashPower( power );    
    myAcceleration.setPolar(dashPowerRate() * power, faceDir() );
    setSpeed( speed()+ myAcceleration );
    if( speed().length() > maxSpeed() ) {
      Vector mySpeed = speed();
      mySpeed.setPolar(maxSpeed(), mySpeed.angle());
      setSpeed(mySpeed);
    }
  }
  else {
    throwException( 
        ModelledPlayerException( "ModelledPlayer has no valid speed value" ) );
  }
}

void ModelledPlayer::turn( double angle ) 
{
  angle = effectiveTurnAngle(angle);
  theFaceDir = normalizeAngle( theFaceDir + angle );
  if ( theFaceDir < 0 ) {
    theFaceDir += 2 * PI;
  }  
}

double ModelledPlayer::angleTo( const Vector& point ) const 
{
  Vector relativePos = point - pos();
  if( faceDir() != UNDEF_DOUBLE ) {
    double angleToObject = relativePos.angle() - faceDir();
    return normalizeAngle(angleToObject);
  }
  else{
    throwException( 
        ModelledPlayerException( "ModelledPlayer has undefined faceDir" ) );
  }
}

bool ModelledPlayer::isInFieldOfView( const StableObject& object ) const 
{
  double angleToObject = angleTo( object.pos() );
  if( fabs(angleToObject) < visibleAngle() / 2.0 ) {
    return YES;
  }
  else {
    return NO;
  }
}

bool ModelledPlayer::isInFieldOfFeel( const StableObject& object ) const
{
  if( (object.pos() - pos()).length() <= globalConf->visibleDistance ) {
    return YES;
  }
  else {
    return NO;
  }
}

bool ModelledPlayer::isInFieldOfSense( const StableObject& object ) const
{
  return ( isInFieldOfView( object ) || isInFieldOfFeel( object ) );
}

double ModelledPlayer::effectiveDashPower( double power ) const
{
  double theEffectivePower;
  
  power = normalizeDashPower( power );
  if ( fabs( power ) > theStamina ) {
    power = signum(power) * theStamina ;
  }
  theEffectivePower = power * theEffort;

  return theEffectivePower;
}

double ModelledPlayer::dashPowerForSend( double power ) const
{
  double thePowerForSend;

  power = normalizeDashPower( power );
  power = power / theEffort;
  thePowerForSend = normalizeDashPower( power );

  if ( fabs( thePowerForSend ) > stamina() ) {
    WARNING( Log::MODELLED_OBJECTS,
             "dashPowerForSend: thePowerForSend > stamina" );
  }
  if( !isAlmostEqual( thePowerForSend, power ) ) {
    WARNING( Log::MODELLED_OBJECTS, 
             "dashPowerForSend: power value may be out of range" );
  }

  return thePowerForSend;
}

double ModelledPlayer::maxEffectiveDashPower() const
{
  INFO( Log::MODELLED_OBJECTS, 4, 
        "ENTER: ModelledPlayer::maxEffectiveDashPower\n" );
  double power;
  Vector acceleration;

  if( speedReliability().get() < Reliability::RANDOM ) {
    power = effectiveDashPower( maxPower() );
    acceleration.setPolar( dashPowerRate() * power, faceDir() );

    if( (speed() + acceleration).length() > maxSpeed() ) {
      power = (maxSpeed() - speed().length()) / dashPowerRate();
      acceleration.setPolar( dashPowerRate() * power, faceDir() );
      INFO( Log::MODELLED_OBJECTS, 15, "Started with power: " + 
            String(power) + " and Speed of: " + 
            String( ( speed() + acceleration ).length() ) + "\n" );
      double oldPower = power;
      while( (speed() + acceleration).length() < maxSpeed() &&
             power < maxPower() ) {
        power++;
        acceleration.setPolar( dashPowerRate() * power, faceDir() );
      }
      if( oldPower < power ) {
        power--;
      }
      if( IS_ACTIVE( Log::MODELLED_OBJECTS, 15 ) ) {
        acceleration.setPolar( dashPowerRate() * power, faceDir() );
        globalLog->info( "Ended with power: " + String(power) + 
            " and Speed of: " + 
            String( ( speed() + acceleration ).length() ) + "\n" );
      }
    }
  }
  else {
    throwException(
        ModelledPlayerException( "ModelledPlayer has invalid speed" ) );
  }

  INFO( Log::MODELLED_OBJECTS, 4, 
        "LEAVE: ModelledPlayer::maxEffectiveDashPower\n" );
  return power;
}

double ModelledPlayer::effectiveTurnAngle( double angle ) const 
{
  double theEffectiveTurnAngle;

  angle = normalizeTurnAngle(angle);
  theEffectiveTurnAngle = angle / turnAngleMult();

  return theEffectiveTurnAngle;
}

double ModelledPlayer::turnAngleForSend( double angle ) const 
{
  double theAngleForSend;

  angle = normalizeAngle( angle );
  angle = angle * turnAngleMult();
  theAngleForSend = normalizeTurnAngle( angle );
  if( !isAlmostEqual( theAngleForSend, angle ) ) {
    WARNING( Log::MODELLED_OBJECTS, 
             "turnAngleForSend: turn angle may be out of range");
  }

  return theAngleForSend;
}

double ModelledPlayer::maxEffectiveTurnAngle() const 
{
  return effectiveTurnAngle( maxMoment() ); 
}

double ModelledPlayer::turnAngleMult() const 
{
  return( 1.0 + inertiaMomentOfTurn() * speed().length() );
}

void ModelledPlayer::recover( int simSteps ) 
{
  while( simSteps > 0 ) {
    /* RecoveryFactor ausrechnen */
    if( theStamina <= recoverDecThr() * staminaMax() ) {
      if( theRecovery <= recoverMin() ){
        /* Minimum erreicht */
      } 
      else{
        theRecovery -= recoverDec();
      }
    }
    
    /* EffortFactor ausrechnen */
    if( theStamina <= effortDecThr() * staminaMax() ) {
      if( theEffort <= effortMin() ) {
        /* Minimum erreicht */
      }
      else{ 
        theEffort -= effortDec();
      }
    }
    if( theStamina >= effortIncThr() * staminaMax() ) {
      if( theEffort >= 1.0 ) {
        /* Maximum erreicht */
      }
      else {
        theEffort += effortInc();
      }
    }

    /* Stamina ausrechnen */
    theStamina += (theRecovery * staminaIncMax());
    if ( theStamina > staminaMax() ){
      theStamina=staminaMax();
    }

    theStaminaReliability.decrease();
    simSteps--;
  }
}


int ModelledPlayer::estimatedStepsToBall( ModelledBall ball ) const 
{
  INFO( Log::MODELLED_OBJECTS, 4, 
        "ENTER ModelledPlayer::estimatedStepsToBall\n" );
  if( distTo( ball.pos() ) > globalConf->distToBallForEvaluation ) {
    INFO( Log::MODELLED_OBJECTS, 4, 
          "LEAVE ModelledPlayer::estimatedStepsToBall\n" );
    return MAX_STEPS_TO_BALL;
  }

  double balldist = ( ball.pos() - pos() ).length();
  double interceptdist = ball.size() + ball.kickMargin() + size() -
    DEFAULT_COMPARISON_ACCURACY;
  double averageSpeed = maxSpeed();
  int steps = 0;
  int maxSteps = (int) ( globalConf->distToBallForEvaluation / averageSpeed );
  double currentStamina = stamina();
  double staminaLimit = max( effortDecThr(), recoverDecThr() ) * staminaMax();

  if( balldist < interceptdist ) {
    INFO( Log::MODELLED_OBJECTS, 4, 
          "LEAVE ModelledPlayer::estimatedStepsToBall\n" );
    return steps;
  }
  interceptdist -= TRIANGULATION_ACCURACY;

  if( currentStamina < staminaLimit ) {
    WARNING( Log::MODELLED_OBJECTS, "stamina less then limit\n");
    steps += (int) ( (staminaLimit - currentStamina) / staminaIncMax() ) + 1;
    currentStamina += steps * staminaIncMax();
  }

  while( balldist > interceptdist && steps < maxSteps ) {
    ball.simulateSteps();
    balldist = ( ball.pos() - pos() ).length();
    interceptdist += averageSpeed;
    steps++;
  }
  INFO( Log::MODELLED_OBJECTS, 9, "calculated Steps before stamina: " +
        String( steps ) + "\n" );

  double estimatedStaminaUse = steps * ( ( averageSpeed / 
      ( globalConf->distPerStaminaPoint * effort() ) ) - 
      staminaIncMax() );
  double staminaDiff = currentStamina - staminaLimit - estimatedStaminaUse;
  if( staminaDiff < 0 ) {
    steps += (int) ( (-staminaDiff) / staminaIncMax() ) + 1;
  }
  if( steps > MAX_STEPS_TO_BALL ) {
    steps = MAX_STEPS_TO_BALL;
  }
  INFO( Log::MODELLED_OBJECTS, 9, "calculated Steps after stamina: " +
        String( steps ) + " with use of " + String( estimatedStaminaUse ) +
        " stamina points.\n" );

  INFO( Log::MODELLED_OBJECTS, 4, 
        "LEAVE ModelledPlayer::estimatedStepsToBall\n" );

  return steps;
}

int ModelledPlayer::calculatedStepsToBall( ModelledBall& ball,
                                          int estimatedSteps ) const
{
  INFO( Log::MODELLED_OBJECTS, 4, 
        "ENTER ModelledPlayer::calculateStepsToBall\n" );  

  ModelledPlayer simulatePlayer = *this;
  simulatePlayer.setId( globalWorldModel->id() );
  simulatePlayer.setSpeedReliability( Reliability::PERFECT );
  if( simulatePlayer.faceDir() == UNDEF_DOUBLE ) {
    simulatePlayer.setFaceDir( 0 );
  }
  Situation scenario = Situation();
  scenario.addPlayer( simulatePlayer );
  double interceptRange = ball.size() + size() + ball.kickMargin() * 
    globalConf->optimalKickDist;
  double staminaLimit = max( effortDecThr(), recoverDecThr() ) * staminaMax();

  int numberOfIterations = 0;
  int lengthOfPlan;
  int steps = estimatedSteps;
  ball.simulateSteps( steps );

  while( YES ) {
    if( steps >= globalConf->maxStepsToCalculate ) {
      INFO( Log::MODELLED_OBJECTS, 7, "Steps to ball was not exact computed," +
            String( "because ball is to far away.\n" ) );
      break;
    }
    if( numberOfIterations >= 5 ) {
      INFO( Log::MODELLED_OBJECTS, 7, "Steps to ball was not exact computed," +
            String( "because of to many iterations." ) ); 
      break;
    }
    globalSkills->resetSituation( scenario );
    lengthOfPlan = globalSkills->gotoRegion( ball.pos(), interceptRange,
        staminaLimit, Skills::STANDARD_BEHAVIOR, 
        steps + globalConf->maxStepsToCalculate ).length();
    numberOfIterations++;
    if( lengthOfPlan > steps ) {
      ball.simulateSteps( lengthOfPlan - steps );
      steps = lengthOfPlan;
    }
    else {
      break;
    }
  }
  
  INFO( Log::MODELLED_OBJECTS, 9, "estimatedSteps: " +
        String( estimatedSteps ) );
  INFO( Log::MODELLED_OBJECTS, 6, "calculatedSteps: " + String( steps ) );
  INFO( Log::MODELLED_OBJECTS, 4, 
        "LEAVE ModelledPlayer::calculateStepsToBall\n" );  
  return steps;
}

int ModelledPlayer::team() const
{
  if ( id() >= MIN_TEAMMATE_ID && id() <= MAX_TEAMMATE_ID ) {
    return OUR_TEAM_ID;
  }
  else if( id() >= MIN_OPPONENT_ID && id() <= MAX_OPPONENT_ID ) {
    return THEIR_TEAM_ID;
  }
  else {
    return UNDEF_TEAM_ID;
  }
}
  
bool ModelledPlayer::isPlayerNumberKnown() const
{
  if( ( id() >= MIN_TEAMMATE_ID && id() <= MAX_EXISTING_TEAMMATE_ID ) ||
      ( id() >= MIN_OPPONENT_ID && id() <= MAX_EXISTING_OPPONENT_ID ) ) {
    return YES;
  }
  else {
    return NO;
  }
}

StorableObject* ModelledPlayer::copy() const 
{
  return new ModelledPlayer( *this );
}


String ModelledPlayer::toString() const 
{
  String output;
 
  output = String( "( ID: " ) + String( id() ) + 
    String(", Pos:") + pos().toString() +
    String( ", Pos" ) + posReliability().toString() +
    String(", Speed:") + speed().toString() +
    String( ", Speed" ) + speedReliability().toString() +
    String(", FaceDir: ") + String(degFromRadiant(faceDir())) + 
    String(", FaceDir" ) + faceDirReliability().toString() +
    String(", VisibleAngle: ") + String(degFromRadiant(visibleAngle())) +
    String(", Stamina: ") + String(stamina()) +
    String(", Recovery: ") + String(recovery()) +
    String(", Effort: ") + String(effort()) +
    String(", StaminaValues" ) + staminaReliability().toString() +
    String( " )" );
  
  return output;
}







